import unittest

from defern import defern, deferner, defern_this, here


class TestDefern(unittest.TestCase):

    @staticmethod
    def add_defern(frame, result):
        defern(result.append, 7, frame=frame)

    def _test_func_1(self, result):
        result.append(1)

        defern(result.append, 3)
        defern(lambda: result.append(4))

        @defern_this
        def append_5():
            result.append(5)

        @deferner
        def append_num(num):
            result.append(num)

        append_num(6)

        self.add_defern(here(), result)

        result.append(2)

    def test_core(self):
        result = []

        self._test_func_1(result)
        print(result)
        self.assertEqual(result, [1, 2, 3, 4, 5, 6, 7])


if __name__ == '__main__':
    unittest.main()



import inspect
import uuid
import warnings
from types import FrameType
from typing import Any, Tuple, Callable
from typing import AnyStr, Optional
from functools import wraps


def _outer_frame(frame: FrameType = None) -> FrameType:
    return frame or inspect.currentframe().f_back.f_back


class Defern:
    def __init__(self, func: Callable, args, kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __del__(self):
        self.func(*self.args, **self.kwargs)


def defern(func: Callable, *args, frame: FrameType = None, **kwargs) -> Tuple[AnyStr, Defern]:
    """
    defern(func, *args, frame: FrameType = None, **kwargs)

    Defer the callback to the end of the frame.

    :param func: The callback function.
    :param args: The arguments to pass to the callback function.
    :param frame: The frame to defer the callback to.
    :param kwargs: The keyword arguments to pass to the callback function.
    """
    warnings.warn("defern(func, *args, frame, **kwargs) is deprecated, use defer(func, frame)(*args, **kwargs) instead",
                  DeprecationWarning)
    deferer = Defern(func, args, kwargs)

    new_variable_name = 'defern-{}'.format(uuid.uuid4())
    frame = _outer_frame(frame)
    frame.f_locals[new_variable_name] = deferer

    return new_variable_name, deferer


def deferner(func: Callable) -> Callable:
    """
    deferner(func)

    Wrap the callback function to defer the callback to the end of the frame.

    :param func: The callback function.
    :return: The callback function.
    """
    warnings.warn(
        'deferner(func)(*args, frame, **kwargs) is deprecated, use defer(func, frame)(*args, **kwargs) instead',
        DeprecationWarning)

    @wraps(func)
    def wrapper(*args, frame: FrameType = None, **kwargs) -> Tuple[AnyStr, Defern]:
        deferer = Defern(func, args, kwargs)

        new_variable_name = 'defern-{}'.format(uuid.uuid4())
        frame = _outer_frame(frame)
        frame.f_locals[new_variable_name] = deferer
        return new_variable_name, deferer

    return wrapper


def defern_this(func: Callable) -> Callable:
    """
    defern_this(func)

    Wrap the callback function to defer the callback to the end of the frame.

    :param func: The callback function.
    :return: The callback function.
    """
    warnings.warn('defern_this(func) is deprecated, use defer(func)() instead', DeprecationWarning)
    deferer = Defern(func, tuple(), dict())

    new_variable_name = 'defern-{}'.format(uuid.uuid4())
    frame = _outer_frame()
    frame.f_locals[new_variable_name] = deferer

    return func


class DeferStack:
    def __init__(self, frame: FrameType):
        self.frame = frame

    @property
    def back(self) -> 'DeferStack':
        return DeferStack(self.frame.f_back)


def here() -> DeferStack:
    return DeferStack(inspect.currentframe().f_back)


class DefernCaller:
    namespace = 'DefernCaller'

    def __init__(self):
        self.callback = []

    def set(self, func: Callable, args: Tuple[Any], kwargs: [str, Any], frame: DeferStack):
        caller_name = frame.frame.f_code.co_name
        caller_locals = frame.frame.f_locals
        defern_id = 'Defern_' + uuid.uuid3(uuid.NAMESPACE_DNS, caller_name).hex.replace('-', '_')
        if defern_id in caller_locals:
            defern_caller = caller_locals[defern_id]
            if defern_caller is not self:
                caller_locals[defern_id].set(func, args, kwargs, frame)
                return
        else:
            caller_locals[defern_id] = self
        self.callback.append((func, args, kwargs))

    def __del__(self):
        while self.callback:
            func, args, kwargs = self.callback.pop()
            func(*args, **kwargs)


def defer(callback, frame: Optional[DeferStack] = None):
    """
    defer(callback, frame: Optional[DeferStack] = None)

    Defer the callback to the end of the frame.

    :param callback: The callback function.
    :param frame: The frame to defer the callback to.
    :return: The callback function.

    """

    @wraps(callback)
    def func(*args, **kwargs):
        nonlocal frame
        if frame is None:
            frame = here().back
        DefernCaller().set(callback, args, kwargs, frame)

    return func



__title__ = "defern"
__description__ = "Provides Go-like `defer`"
__url__ = "https://github.com/nagataaaas/defern"
__version_info__ = ("1", "0", "0")
__version__ = ".".join(__version_info__)
__author__ = "Yamato Nagata"
__author_email__ = "nagata@mail.nagata.pro"
__maintainer__ = "Yamato Nagata"
__license__ = "MIT"
__copyright__ = "(c) 2021-2022 by Yamato Nagata"



from .__about__ import __version__
from .defern import (defern, deferner, defern_this, here, defer)


__all__ = [
    __version__,
    "defern",
    "deferner",
    "defern_this",
    "here",
    "defer",
]



"""
Provides Go-like `defer`
-----------
Powered by [Yamato Nagata](https://twitter.com/514YJ)

[GitHub](https://github.com/nagataaaas/defern)

installation:
`$ pip install defern`

```python
from defern import defern, deferner, defern_this, here

def defer_multiple_function(frame):
    defern(lambda: print("it's middle 7th"), frame=frame)
    defern(lambda: print("it's middle 8th"), frame=frame)

print("Hi")

defern(lambda: print("it's middle 1st"))  # you can pass function to run after return.
defern(lambda: print("it's middle 2nd"))
defern(print, "it's middle 3rd")  # passed args and kwargs after function will given to function.

@deferner
def defer_this(number: str):
    print("it's middle", number)  # you can create function which runs after return with `@deferner`

defer_this("4th")
defer_this("5th", frame=here())

@defern_this  # wrap function to create procedure which automatically runs after return
def defer_this_now():
    print("it's middle 6th")

defer_multiple_function(here())  # here() to get currentFrame and pass it to `defern` or function created with `deferner`
                                 # with name `frame` will create `defern` to that frame

print('ended')

# Hi
# ended
# it's middle 1st
# it's middle 2nd
# it's middle 3rd
# it's middle 4th
# it's middle 5th
# it's middle 6th
# it's middle 7th
# it's middle 8th
```
"""

from setuptools import setup
from os import path

about = {}
with open("defern/__about__.py") as f:
    exec(f.read(), about)

here = path.abspath(path.dirname(__file__))

setup(name=about["__title__"],
      version=about["__version__"],
      url=about["__url__"],
      license=about["__license__"],
      author=about["__author__"],
      author_email=about["__author_email__"],
      description=about["__description__"],
      long_description=__doc__,
      long_description_content_type="text/markdown",
      packages=["defern"],
      zip_safe=False,
      platforms="any",
      classifiers=[
          "Development Status :: 4 - Beta",
          "Environment :: Other Environment",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python",
          "Topic :: Software Development :: Libraries :: Python Modules"
      ])
